"""
audit
=====

This package manages audits for a given model. An audit can consist of different types:
- Training
- Production (or Testing)
- Recorded (or Impacted)

Each audit class captures logic and data relevant to that auditing phase.
"""
